# cython: language_level=3
from .common cimport *
from .transducer_header cimport TransducerHeader
from .transducer_alphabet cimport TransducerAlphabet


cdef class Transducer:
    cdef public TransducerHeader header
    cdef public TransducerAlphabet alphabet
    cdef public bint is_weighted
    cdef public dict operations
    cdef public dict symbol_map
    cdef public IndexTable index_table
    cdef public TransitionTable transition_table
    cpdef void construct_symbol_map(self)


# cython: language_level=3
from .common cimport *
from .flag_diacritic_operator cimport FlagDiacriticOperator

cdef class FlagDiacriticOperation:
    """
    Class representing a flag diacritic operation.
    """
    def __init__(self, FlagDiacriticOperator operation, int feat, int val=0):
        """
        Initializes the FlagDiacriticOperation instance.

        :param operation: The flag diacritic operator. Defaults to None.
        :param feat: The feature index. Defaults to None.
        :param val: The value of the feature. Defaults to None.

        """
        self.op = operation if operation else FlagDiacriticOperator.P
        self.feature = feat
        self.value = val

    cpdef bint is_flag(self):
        """
        Checks if the operation is a flag diacritic operation.

        :return: True if the operation is a flag diacritic operation, False otherwise.
        """
        return self.feature != NO_SYMBOL_NUMBER



from typing import Union, List, Tuple
from pathlib import Path
from io import BufferedReader


try:
    from c_pyhfst.transducer import Transducer
    from c_pyhfst.transducer_header import TransducerHeader
    from c_pyhfst.transducer_alphabet import TransducerAlphabet
    from c_pyhfst.analyzer import Analyzer
except:
    from .transducer import Transducer
    from .transducer_header import TransducerHeader
    from .transducer_alphabet import TransducerAlphabet
    from .analyzer import Analyzer

def get_transducer(transducer_path: Union[str, Path]) -> Transducer:
    """
    Creates a Transducer instance from the given transducer file path.

    :param transducer_path: The path to the transducer file.
    :return: A Transducer instance.
    """
    with open(transducer_path, "rb") as transducer_file:
        char_stream = BufferedReader(transducer_file)
        transducer_header = TransducerHeader(char_stream)
        transducer_alphabet = TransducerAlphabet(
            char_stream, transducer_header.get_symbol_count()
        )
        return Transducer(char_stream, transducer_header, transducer_alphabet, is_weighted=transducer_header.is_weighted())


class HfstInputStream(object):
    def __init__(self, path: Union[str, Path], cache=True) -> None:
        """
        Initialize an HfstInputStream object.

        :param path: The path to the transducer file.
        :param cache: Whether to cache the results.
        """
        self.path = path
        self.cache = cache

    def read(self) -> 'Hfst':
        """
        Read the transducer file and return an Hfst object.

        :return: An Hfst object initialized with the transducer read from the file.
        """
        tr = get_transducer(self.path)
        return Hfst(tr, cache=self.cache)


class Hfst(object):
    def __init__(self, tr: Transducer, cache=True) -> None:
        """
        Initialize an Hfst object with a given transducer.

        :param tr: The transducer object.
        :param cache: Whether to cache the results.
        """
        self.tr = tr
        self.cache = cache
        self.mem = {}


    def lookup(self, string: str) -> List[Tuple[str, float]]:
        """
        Perform lookup on the input string and return the analyses.

        :param string: The input string to analyze.
        :return: A list of tuples, where each sublist contains the string representation of the result and its weight.
        """
        if self.cache:
            if string not in self.mem:
                self.mem[string] = Analyzer(self.tr, string).analyze()
            result = self.mem[string]
        else:
            result = Analyzer(self.tr, string).analyze()
        return [["".join(_r.get_symbols()), _r.get_weight()] for _r in result]



# cython: language_level=3
from libc.stdint cimport uint8_t, uint16_t, uint32_t
import struct

cdef class ByteArray:
    """
    A simple class for handling byte arrays with various read operations.
    """
    def __init__(self, int s=0, ByteArray another=None):
        if another is not None:
            s = max(s, another.get_size())
            self.bytes = bytearray(another.get_bytes())
        else:
            self.bytes = bytearray(s)
        self.size = s
        self.index = 0

    cpdef public int get_size(self):
        """
        Returns the size of the byte array.

        :return: The size of the byte array.
        """
        return self.size

    cpdef public bytes get(self, int i):
        """
        Returns the byte at the given index.

        :param i: The index of the byte to retrieve.
        :return: The byte at the specified index.
        """
        return self.bytes[i]

    cpdef public bytearray get_bytes(self):
        """
        Returns the byte array.

        :return: The byte array.
        """
        return self.bytes

    cpdef public uint8_t get_ubyte(self):
        """
        Returns an unsigned byte from the byte array and advances the index.

        :return: An unsigned byte.
        """
        cdef uint8_t result = self.bytes[self.index]
        self.index += 1
        return result

    cpdef public uint16_t get_ushort(self):
        """
        Returns an unsigned short from the byte array and advances the index.

        :return: An unsigned short.
        """
        cdef uint16_t result = struct.unpack_from("<H", self.bytes, self.index)[0]
        self.index += 2
        return result

    cpdef public uint32_t get_uint(self):
        """
        Returns an unsigned integer from the byte array and advances the index.

        :return: An unsigned integer.
        """
        cdef uint32_t result = struct.unpack_from("<I", self.bytes, self.index)[0]
        self.index += 4
        return result

    cpdef public bint get_bool(self):
        """
        Returns a boolean value from the byte array and advances the index.

        :return: A boolean value.
        """
        return self.get_uint() != 0

    cpdef public float get_float(self):
        """
        Returns a floating-point number from the byte array and advances the index.

        :return: A floating-point number.
        """
        cdef float result = struct.unpack_from("<f", self.bytes, self.index)[0]
        self.index += 4
        return result



# cython: language_level=3
from .common cimport *
from .transducer_header cimport TransducerHeader
from .transducer_alphabet cimport TransducerAlphabet

cdef class Transducer:
    """
    A class representing a finite state transducer for morphological analysis.
    """

    def __init__(self, file, TransducerHeader h, TransducerAlphabet a, bint is_weighted=True):
        """
        Initializes the Transducer instance.

        :param file: A file containing the transducer data.
        :param h: A TransducerHeader instance representing the header of the transducer.
        :param a: A TransducerAlphabet instance representing the alphabet of the transducer.
        :param is_weighted: A boolean indicating if the transducer is weighted. Defaults to True.
        """
        self.header = h
        self.alphabet = a
        self.is_weighted = is_weighted
        self.operations = self.alphabet.operations
        self.symbol_map = {}
        self.construct_symbol_map()
        self.index_table = IndexTable(file, h.get_index_table_size())
        self.transition_table = TransitionTable(
            file, h.get_target_table_size(), is_weighted=self.is_weighted)

    cpdef void construct_symbol_map(self):
        for i in range(self.header.get_input_symbol_count()):
            _w = self.alphabet.keyTable[i]
            if len(_w) <= 1:
                self.symbol_map[_w] = {None: i}
            else:
                if _w[0] not in self.symbol_map:
                    self.symbol_map[_w[0]] = {}
                _o = self.symbol_map[_w[0]]
                for j in range(1, len(_w)):
                    _c = _w[j]
                    if _c not in _o:
                        _o[_c] = {}
                    _o = _o[_c]
                if None in _o:
                    raise ValueError("Duplicate symbol in symbol map")
                _o[None] = i


# cython: language_level=3
from libc.stdint cimport uint32_t, uint16_t
from cpython cimport dict, array
import cython
from .transducer cimport Transducer

cdef extern from "constants.h":   
    cdef cython.longlong TRANSITION_TARGET_TABLE_START = 2147483648
    cdef cython.longlong INFINITE_WEIGHT = 4294967295
    cdef cython.longlong NO_SYMBOL_NUMBER = 65535
    cdef cython.longlong NO_TABLE_INDEX = 4294967295

cdef class IndexTable:
    cdef array.array ti_input_symbols
    cdef array.array ti_targets
    cdef cython.longlong _size

    cpdef uint16_t get_input(self, int i)
    cpdef uint32_t get_target(self, int i)
    cpdef bint is_final(self, int i)
    cpdef float get_final_weight(self, int i)

cdef class TransitionTable:
    cdef array.array ti_input_symbols
    cdef array.array ti_output_symbols
    cdef array.array ti_targets
    cdef array.array ti_weights
    cdef bint is_weighted
    cdef cython.longlong _size

    cpdef uint16_t get_input(self, int pos)
    cpdef uint16_t get_output(self, int pos)
    cpdef uint32_t get_target(self, int pos)
    cpdef float get_weight(self, int pos)
    cpdef bint is_final(self, int pos)
    cpdef int size(self)

cdef class State:
    cdef public Transducer parent
    cdef public list state_stack
    cdef public list output_string
    cdef public list input_string
    cdef public int output_pointer
    cdef public int input_pointer
    cdef public float current_weight
    cdef public list display_vector

    cdef list find_key(self, str index_string)
    
cdef class Result:
    cdef list symbols
    cdef float weight

    cpdef list get_symbols(self)
    cpdef float get_weight(self)



# cython: language_level=3
cdef class TransducerAlphabet:
    cdef public list keyTable
    cdef public dict operations
    cdef public dict feature_bucket
    cdef public dict value_bucket
    cdef public int features
    cdef public int values


from enum import Enum
from .common import *


class FlagDiacriticOperator(Enum):
    """
    Enum representing flag diacritic operators.
    """
    P = 0
    N = 1
    R = 2
    D = 3
    C = 4
    U = 5


class FlagDiacriticOperation:
    """
    Class representing a flag diacritic operation.
    """

    def __init__(self, operation: Optional[FlagDiacriticOperator] = None, feat: Optional[int] = None, val: Optional[int] = None):
        """
        Initializes the FlagDiacriticOperation instance.

        :param operation: The flag diacritic operator. Defaults to None.
        :param feat: The feature index. Defaults to None.
        :param val: The value of the feature. Defaults to None.

        """
        if operation is not None:
            self.op = operation
            self.feature = feat
            self.value = val
        else:
            self.op = FlagDiacriticOperator.P
            self.feature = NO_SYMBOL_NUMBER
            self.value = 0

    def is_flag(self) -> bool:
        """
        Checks if the operation is a flag diacritic operation.

        :return: True if the operation is a flag diacritic operation, False otherwise.
        """
        return self.feature != NO_SYMBOL_NUMBER



from typing import List, Generator, Any, Tuple, Union, Optional
import io
from collections.abc import ByteString
from collections import defaultdict
from .byte_array import ByteArray


TRANSITION_TARGET_TABLE_START = 2147483648  # 2^31 or UINT_MAX/2 rounded up
# this is hopefully the same as static_cast<float>(UINT_MAX) in C++
INFINITE_WEIGHT = float(4294967295)
NO_SYMBOL_NUMBER = 65535  # USHRT_MAX
NO_TABLE_INDEX = 4294967295


class IndexTable:
    """
    A table to store input symbols and their corresponding target indices.
    """

    def __init__(self, input_stream: io.BytesIO, indices_count: int):
        b = ByteArray(indices_count * 6)
        input_stream.readinto(b.bytes)
        self.ti_input_symbols: List[int] = [0] * indices_count
        self.ti_targets: List[int] = [0] * indices_count
        for i in range(indices_count):
            self.ti_input_symbols[i] = b.get_ushort()
            self.ti_targets[i] = b.get_uint()

    def get_input(self, i: int) -> int:
        """
        Returns the input symbol at the given index.

        :param i: The index to retrieve the input symbol from.
        :return: The input symbol at the specified index.
        """
        return self.ti_input_symbols[i]

    def get_target(self, i: int) -> int:
        """
        Returns the target index at the given index.

        :param i: The index to retrieve the target index from.
        :return: The target index at the specified index.
        """
        return self.ti_targets[i]

    def is_final(self, i: int) -> bool:
        """
        Checks if the given index is a final state.

        :param i: The index to check.
        :return: True if the index is a final state, False otherwise.
        """
        return (
            self.ti_input_symbols[i] == NO_SYMBOL_NUMBER
            and self.ti_targets[i] != NO_TABLE_INDEX
        )

    def get_final_weight(self, i: int) -> float:
        """
        Returns the final weight for the given index.

        :param i: The index to retrieve the final weight from.
        :return: The final weight at the specified index.
        """
        return float(self.ti_targets[i])


class TransitionTable:
    """
    A table to store transitions between states.
    """

    def __init__(
        self, input_stream: io.BytesIO, transition_count: int, is_weighted: bool = True
    ):
        block_size = 12 if is_weighted else 8
        b = ByteArray(transition_count * block_size)
        input_stream.readinto(b.bytes)
        self.is_weighted: bool = is_weighted
        self.ti_input_symbols: List[int] = [0] * transition_count
        self.ti_output_symbols: List[int] = [0] * transition_count
        self.ti_targets: List[int] = [0] * transition_count
        self.ti_weights: List[float] = [0.0] * transition_count

        for i in range(transition_count):
            self.ti_input_symbols[i] = b.get_ushort()
            self.ti_output_symbols[i] = b.get_ushort()
            self.ti_targets[i] = b.get_uint()
            if self.is_weighted:
                self.ti_weights[i] = b.get_float()

    def get_input(self, pos: int) -> int:
        """
        Returns the input symbol at the given position.

        :param pos: The position to retrieve the input symbol from.
        :return: The input symbol at the specified position.
        """
        return self.ti_input_symbols[pos]

    def get_output(self, pos: int) -> int:
        """
        Returns the output symbol at the given position.

        :param pos: The position to retrieve the output symbol from.
        :return: The output symbol at the specified position.
        """
        return self.ti_output_symbols[pos]

    def get_target(self, pos: int) -> int:
        """
        Returns the target index at the given position.

        :param pos: The position to retrieve the target index from.
        :return: The target index at the specified position.
        """
        return self.ti_targets[pos]

    def get_weight(self, pos: int) -> float:
        """
        Returns the weight at the given position.

        :param pos: The position to retrieve the weight from.
        :return: The weight at the specified position.
        :raises Exception: If the transition table is unweighted.
        """
        if not self.is_weighted:
            raise Exception("Getting weights of unweighted FST.")
        return self.ti_weights[pos]

    def is_final(self, pos: int) -> bool:
        """
        Checks if the given position is a final state.
        :param pos: The position to check.
        :return: True if the position is a final state, False otherwise.
        """
        return (
            self.ti_input_symbols[pos] == NO_SYMBOL_NUMBER
            and self.ti_output_symbols[pos] == NO_SYMBOL_NUMBER
            and self.ti_targets[pos] == 1
        )

    def size(self) -> int:
        """
        Returns the size of the transition table.

        :return: The size of the transition table.
        """
        return len(self.ti_targets)


class State:
    """
    A class representing the state of the FST.
    """

    def __init__(self, input: str, parent: Any):
        self.parent: Any = parent
        self.state_stack: List[List[int]] = []
        neutral: List[int] = [0] * parent.alphabet.features
        self.state_stack.append(neutral)
        self.output_string: List[int] = [NO_SYMBOL_NUMBER] * 1000
        self.input_string: List[int] = list(self.find_key(input))
        self.output_pointer: int = 0
        self.input_pointer: int = 0
        self.current_weight: float = 0.0
        self.display_vector: List[int] = []

    def find_key(self, index_string: str) -> Generator[int, None, None]:
        """
        Finds the key associated with the given index string.
        :param index_string: The index string to search for.
        :return: The symbol number if found, otherwise NO_SYMBOL_NUMBER.
        """

        i = 0

        while i < len(index_string):
            match_found = False
            for length in range(len(index_string) - i, 0, -1):
                substr = index_string[i : i + length]
                map_pointer = self.parent.symbol_map
                for char in substr:
                    if char in map_pointer:
                        map_pointer = map_pointer[char]
                    else:
                        break
                else:
                    if None in map_pointer:
                        yield map_pointer[None]
                        i += length - 1
                        match_found = True
                    break
            if not match_found:
                yield NO_SYMBOL_NUMBER
            i += 1

        yield NO_SYMBOL_NUMBER


class Result:
    """
    A class representing a result with symbols and a weight.
    """

    def __init__(self, symbols: List[str], weight: float):
        self.symbols: List[str] = symbols
        self.weight: float = weight

    def get_symbols(self) -> List[str]:
        """
        Returns the list of symbols in the result.

        :return: The list of symbols.
        """
        return self.symbols

    def get_weight(self) -> float:
        """
        Returns the weight associated with the result.

        :return: The weight of the result.
        """
        return self.weight

    def __str__(self) -> str:
        """
        Returns a string representation of the result.

        :return: A string representation of the result in the format (text: weight).
        """
        return "".join(self.symbols) + ": " + str(self.weight)



# cython: language_level=3
cdef class TransducerHeader:
    cdef int number_of_input_symbols
    cdef int number_of_symbols
    cdef int size_of_transition_index_table
    cdef int size_of_transition_target_table
    cdef int number_of_states
    cdef int number_of_transitions
    cdef bint weighted
    cdef bint deterministic
    cdef bint input_deterministic
    cdef bint minimized
    cdef bint cyclic
    cdef bint has_epsilon_epsilon_transitions
    cdef bint has_input_epsilon_transitions
    cdef bint has_input_epsilon_cycles
    cdef bint has_unweighted_input_epsilon_cycles
    cdef bint hfst3
    
    cpdef bint begins_hfst3_header(self, bytes b)
    cpdef public int get_input_symbol_count(self)
    cpdef public int get_symbol_count(self)
    cpdef public int get_index_table_size(self)
    cpdef public int get_target_table_size(self)
    cpdef public bint is_weighted(self)
    cpdef bint has_hfst3_header(self)



import io
import struct
from .byte_array import ByteArray


class TransducerHeader:
    """
    A class representing the header of the transducer.
    """

    def __init__(self, input_bytes: io.BytesIO):
        read_bytes: bytes = input_bytes.read(5)  # "HFST\0"
        if self.begins_hfst3_header(read_bytes):
            # just ignore any hfst3 header
            self.hfst3: bool = True
            remaining = struct.unpack_from("<H", input_bytes.read(3), 0)[
                0]  # get remaining
            struct.unpack_from(f"<{remaining}c", input_bytes.read(
                remaining), 0)  # skip remaining
            # 2 unsigned shorts, 4 unsigned ints and 9 uint-bools
            read_bytes = input_bytes.read(56)
        else:
            read_bytes = read_bytes + input_bytes.read(56 - 5)

        bytes_array = ByteArray(len(read_bytes))
        bytes_array.bytes = read_bytes

        self.number_of_input_symbols = bytes_array.get_ushort()
        self.number_of_symbols = bytes_array.get_ushort()
        self.size_of_transition_index_table = bytes_array.get_uint()
        self.size_of_transition_target_table = bytes_array.get_uint()
        self.number_of_states = bytes_array.get_uint()
        self.number_of_transitions = bytes_array.get_uint()
        self.weighted = bytes_array.get_bool()
        self.deterministic = bytes_array.get_bool()
        self.input_deterministic = bytes_array.get_bool()
        self.minimized = bytes_array.get_bool()
        self.cyclic = bytes_array.get_bool()
        self.has_epsilon_epsilon_transitions = bytes_array.get_bool()
        self.has_input_epsilon_transitions = bytes_array.get_bool()
        self.has_input_epsilon_cycles = bytes_array.get_bool()
        self.has_unweighted_input_epsilon_cycles = bytes_array.get_bool()

    def begins_hfst3_header(self, bytes: bytes) -> bool:
        """
        Checks if the given bytes begin with the HFST3 header.

        :param bytes: The bytes to check.
        :return: True if the bytes begin with the HFST3 header, False otherwise.
        """
        return len(bytes) >= 5 and bytes == b"HFST\x00"

    def get_input_symbol_count(self) -> int:
        """
        Returns the number of input symbols.

        :return: The number of input symbols.
        """
        return self.number_of_input_symbols

    def get_symbol_count(self) -> int:
        """
        Returns the number of symbols.

        :return: The number of symbols.
        """
        return self.number_of_symbols

    def get_index_table_size(self) -> int:
        """
        Returns the size of the transition index table.
        :return: The size of the transition index table.
        """
        return self.size_of_transition_index_table

    def get_target_table_size(self) -> int:
        """
        Returns the size of the transition target table.

        :return: The size of the transition target table.
        """
        return self.size_of_transition_target_table

    def is_weighted(self) -> bool:
        """
        Checks if the transducer is weighted.

        :return: True if the transducer is weighted, False otherwise.
        """
        return self.weighted

    def has_hfst3_header(self) -> bool:
        """
        Checks if the transducer has an HFST3 header.

        :return: True if the transducer has an HFST3 header, False otherwise.
        """
        return self.hfst3



from typing import Union, List
from pathlib import Path
from io import BufferedReader
from .transducer_header cimport TransducerHeader
from .transducer_alphabet cimport TransducerAlphabet
from .transducer cimport Transducer as Transducer
from .analyzer import Analyzer

cpdef Transducer get_transducer(transducer_path: Union[str, Path]):
    """
    Creates a Transducer instance from the given transducer file path.

    :param transducer_path: The path to the transducer file.
    :return: A Transducer instance.
    """
    with open(transducer_path, "rb") as transducer_file:
        char_stream = BufferedReader(transducer_file)
        transducer_header = TransducerHeader(char_stream)
        transducer_alphabet = TransducerAlphabet(
            char_stream, transducer_header.get_symbol_count()
        )
        return Transducer(char_stream, transducer_header, transducer_alphabet, is_weighted=transducer_header.is_weighted())


# cython: language_level=3
from .common cimport *
from .flag_diacritic_operator cimport FlagDiacriticOperator

cdef class FlagDiacriticOperation:
    cdef FlagDiacriticOperator op
    cdef int feature
    cdef int value
    cpdef bint is_flag(self)



# cython: language_level=3
from .common cimport *
from .flag_diacritic_operation cimport FlagDiacriticOperation
from .transducer import Transducer
import cython


cdef class Analyzer:
    cdef Transducer transducer
    cdef str input_str
    cdef State state


    cpdef cython.longlong pivot(self, cython.longlong i)
    cpdef void try_epsilon_indices(self, cython.longlong index)
    cpdef void try_epsilon_transitions(self, cython.longlong index)
    cpdef void find_index(self, cython.longlong index)
    cpdef void handle_epsilon_transition(self, cython.longlong index)
    cpdef void find_transitions(self, cython.longlong index)
    cpdef void get_analyses(self, cython.longlong idx)
    cpdef void update_output_string(self, cython.longlong output_symbol)
    cpdef void handle_end_of_input_string(self, cython.longlong index, bint is_transition)
    cpdef void reset_output_pointer(self)
    cpdef tuple get_final_and_weight(self, cython.longlong index, bint is_transition)
    cpdef void update_and_note_analysis(self, float weight)
    cpdef list get_symbols(self)
    cpdef void note_analysis(self)
    cpdef list get_alphabet(self)
    cpdef public list analyze(self)
    cpdef bint push_state(self, FlagDiacriticOperation flag)


# cython: language_level=3
from .common cimport *
from .flag_diacritic_operation cimport FlagDiacriticOperation, FlagDiacriticOperator
from .transducer import Transducer

cdef class Analyzer:

    def __cinit__(self, transducer: Transducer, input_str: str):
        self.transducer = transducer
        self.input_str = input_str
        self.state = State(input_str, self.transducer)

    cpdef cython.longlong pivot(self, cython.longlong i):
        """
        Computes the pivot for the given index.

        :param i: The index to compute the pivot for.
        :return: The pivot value.
        """
        if i >= TRANSITION_TARGET_TABLE_START:
            return (i - TRANSITION_TARGET_TABLE_START) % TRANSITION_TARGET_TABLE_START
        return i

    cpdef void try_epsilon_indices(self, cython.longlong index):
        """
        Tries epsilon indices for the given index and state.

        :param index: The index to try epsilon indices for.
        :param state: The current state of the transducer.
        """
        if self.transducer.index_table.get_input(index) == 0:
            self.try_epsilon_transitions(self.pivot(
                self.transducer.index_table.get_target(index)))

    cpdef void try_epsilon_transitions(self, cython.longlong index):
        """
        Tries epsilon transitions for the given index and state.

        :param index: The index to try epsilon transitions for.
        :param state: The current state of the transducer.
        """
        cdef cython.longlong input_symbol

        while True:
            input_symbol = self.transducer.transition_table.get_input(index)
            if self.transducer.operations.get(input_symbol):
                if not self.push_state(self.transducer.operations[input_symbol]):
                    index += 1
                    continue
                self.handle_epsilon_transition(index)
                index += 1
                self.state.state_stack.pop()
                continue
            elif input_symbol == 0:
                self.handle_epsilon_transition(index)
                index += 1
                continue
            else:
                break

    cpdef void find_index(self, cython.longlong index):
        """
        Finds the index in the transducer for the given index and state.

        :param index: The index to find in the transducer.
        :param state: The current state of the transducer.
        """
        if self.transducer.index_table.get_input(index + (self.state.input_string[self.state.input_pointer - 1])) == self.state.input_string[self.state.input_pointer - 1]:
            self.find_transitions(self.pivot(self.transducer.index_table.get_target(
                index + self.state.input_string[self.state.input_pointer - 1])))

    cpdef void handle_epsilon_transition(self, cython.longlong index):
        """
        Handles epsilon transitions for the given index and state.

        :param index: The index to handle epsilon transitions for.
        :param state: The current state of the transducer.
        """
        self.update_output_string(self.transducer.transition_table.get_output(index))
        self.state.output_pointer += 1

        if self.transducer.is_weighted:
            self.state.current_weight += self.transducer.transition_table.get_weight(index)

        self.get_analyses(self.transducer.transition_table.get_target(index))

        if self.transducer.is_weighted:
            self.state.current_weight -= self.transducer.transition_table.get_weight(index)

        self.state.output_pointer -= 1

    cpdef void find_transitions(self, cython.longlong index):
        """
        Finds the transitions in the transducer for the given index and state.

        :param index: The index to find transitions for.
        :param state: The current state of the transducer.
        """
        for idx in range(index, self.transducer.transition_table.size()):
            input_symbol = self.transducer.transition_table.get_input(idx)
            if input_symbol == NO_SYMBOL_NUMBER:
                break

            if input_symbol == self.state.input_string[self.state.input_pointer - 1]:
                self.update_output_string(self.transducer.transition_table.get_output(idx))
                self.state.output_pointer += 1

                if self.transducer.is_weighted:
                    self.state.current_weight += self.transducer.transition_table.get_weight(
                        idx)

                self.get_analyses(self.transducer.transition_table.get_target(idx))

                if self.transducer.is_weighted:
                    self.state.current_weight -= self.transducer.transition_table.get_weight(
                        idx)

                self.state.output_pointer -= 1
            else:
                break

    cpdef void get_analyses(self, cython.longlong idx):
        """
        Gets the analyses for the given index and state.

        :param idx: The index to get the analyses for.
        :param state: The current state of the transducer.
        """
        cdef cython.longlong index = self.pivot(idx)
        cdef cython.longlong next_v = index
        is_transition = idx >= TRANSITION_TARGET_TABLE_START
        if is_transition:
            if self.transducer.is_weighted:
                next_v = self.pivot(index)
            self.try_epsilon_transitions(next_v + 1)
        else:
            self.try_epsilon_indices(index + 1)

        if self.state.input_string[self.state.input_pointer] == NO_SYMBOL_NUMBER:
            self.handle_end_of_input_string(index, is_transition)
            return

        self.state.input_pointer += 1

        if is_transition:
            self.find_transitions(index + 1)
        else:
            self.find_index(index + 1)

        self.state.input_pointer -= 1
        self.reset_output_pointer()

    cpdef void update_output_string(self, cython.longlong output_symbol):
        """
        Updates the output string in the state based on the output symbol.

        :param state: The current state of the transducer.
        :param output_symbol: The output symbol to append or update in the output string.
        """
        if self.state.output_pointer == len(self.state.output_string):
            self.state.output_string.append(output_symbol)
        else:
            self.state.output_string[self.state.output_pointer] = output_symbol

    cpdef void handle_end_of_input_string(self, cython.longlong index, bint is_transition):
        """
        Handles the end of the input string.

        :param state: The current state of the transducer.
        :param index: The index being processed.
        :param is_transition: A boolean indicating whether it's a transition table or an index table.
        """
        self.reset_output_pointer()

        is_final, weight = self.get_final_and_weight(index, is_transition)
        if is_final:
            self.update_and_note_analysis(weight)

    cpdef void reset_output_pointer(self):
        """
        Resets the output pointer in the state.

        :param state: The current state of the transducer.
        """
        if self.state.output_pointer == len(self.state.output_string):
            self.state.output_string.append(NO_SYMBOL_NUMBER)
        else:
            self.state.output_string[self.state.output_pointer] = NO_SYMBOL_NUMBER

    cpdef tuple get_final_and_weight(self, cython.longlong index, bint is_transition):
        """
        Gets the final state and weight based on the index and transition flag.

        :param index: The index being processed.
        :param is_transition: A boolean indicating whether it's a transition table or an index table.
        :return: A tuple containing a boolean for final state and a float for the weight.
        """
        if is_transition:
            is_final = self.transducer.transition_table.size(
            ) > index and self.transducer.transition_table.is_final(index)
            weight = self.transducer.transition_table.get_weight(
                index) if self.transducer.is_weighted else 0
        else:
            is_final = self.transducer.index_table.is_final(index)
            weight = self.transducer.index_table.get_final_weight(
                index) if self.transducer.is_weighted else 0

        return is_final, weight

    cpdef void update_and_note_analysis(self, float weight):
        """
        Updates the state's current weight and notes the analysis.

        :param state: The current state of the transducer.
        :param weight: The weight to be
        """
        if self.transducer.is_weighted:
            self.state.current_weight += weight
        self.note_analysis()
        if self.transducer.is_weighted:
            self.state.current_weight -= weight

    cpdef list get_symbols(self):
        """
        Gets the symbols for the given state.

        :param state: The current state of the transducer.
        :return: A list of symbols.
        """
        symbols = [self.transducer.alphabet.keyTable[self.state.output_string[i]] for i in range(
            len(self.state.output_string)) if self.state.output_string[i] != NO_SYMBOL_NUMBER]
        return symbols

    cpdef void note_analysis(self):
        """
        Notes the analysis for the given state.

        :param state: The current state of the transducer.
        """
        self.state.display_vector.append(Result(self.get_symbols(), self.state.current_weight if self.transducer.is_weighted else 1.0))

    cpdef list get_alphabet(self):
        """
        Gets the alphabet of the transducer.

        :return: A list of alphabet symbols.
        """
        return self.transducer.alphabet.keyTable

    cpdef public list analyze(self):
        if self.state.input_string[0] == NO_SYMBOL_NUMBER:
            return []
        else:
            self.get_analyses(0)
            return self.state.display_vector

    cpdef bint push_state(self, FlagDiacriticOperation flag):
        cdef list top = self.state.state_stack[-1].copy()
        if flag.op == FlagDiacriticOperator.P:  # positive set
            self.state.state_stack.append(top)
            self.state.state_stack[-1][flag.feature] = flag.value
            return True
        elif flag.op == FlagDiacriticOperator.N:  # negative set
            self.state.state_stack.append(top)
            self.state.state_stack[-1][flag.feature] = -1 * flag.value
            return True
        elif flag.op == FlagDiacriticOperator.R:  # require
            if flag.value == 0:  # empty require
                if self.state.state_stack[-1][flag.feature] == 0:
                    return False
                else:
                    self.state.state_stack.append(top)
                    return True
            elif self.state.state_stack[-1][flag.feature] == flag.value:
                self.state.state_stack.append(top)
                return True
            return False
        elif flag.op == FlagDiacriticOperator.D:  # disallow
            if flag.value == 0:  # empty disallow
                if self.state.state_stack[-1][flag.feature] != 0:
                    return False
                else:
                    self.state.state_stack.append(top)
                    return True
            elif self.state.state_stack[-1][flag.feature] == flag.value:
                return False
            else:
                self.state.state_stack.append(top)
                return True
        elif flag.op == FlagDiacriticOperator.C:  # clear
            self.state.state_stack.append(top)
            self.state.state_stack[-1][flag.feature] = 0
            return True
        elif flag.op == FlagDiacriticOperator.U:  # unification
            if (self.state.state_stack[-1][flag.feature] == 0) or (self.state.state_stack[-1][flag.feature] == flag.value) or (self.state.state_stack[-1][flag.feature] != flag.value and self.state.state_stack[-1][flag.feature] < 0):
                self.state.state_stack.append(top)
                self.state.state_stack[-1][flag.feature] = flag.value
                return True
            return False
        return False  # compiler sanity



import struct
from typing import Optional


class ByteArray:
    """
    A simple class for handling byte arrays with various read operations.
    """

    def __init__(self, s: Optional[int] = None, another: Optional['ByteArray'] = None):
        if another is not None:
            s = max(s.get_size(), another.get_size())
            self.bytes = bytearray(another.get_bytes())
        else:
            self.bytes = bytearray(s)
        self.size: int = s
        self.index: int = 0

    def get_size(self) -> int:
        """
        Returns the size of the byte array.

        :return: The size of the byte array.
        """
        return self.size

    def get(self, i: int) -> int:
        """
        Returns the byte at the given index.

        :param i: The index of the byte to retrieve.
        :return: The byte at the specified index.
        """
        return self.bytes[i]

    def get_bytes(self) -> bytearray:
        """
        Returns the byte array.

        :return: The byte array.
        """
        return self.bytes

    def get_ubyte(self) -> int:
        """
        Returns an unsigned byte from the byte array and advances the index.

        :return: An unsigned byte.
        """
        result = self.bytes[self.index]
        self.index += 1
        return result

    def get_ushort(self) -> int:
        """
        Returns an unsigned short from the byte array and advances the index.

        :return: An unsigned short.
        """
        result = struct.unpack_from("<H", self.bytes, self.index)[0]
        self.index += 2
        return result

    def get_uint(self) -> int:
        """
        Returns an unsigned integer from the byte array and advances the index.

        :return: An unsigned integer.
        """
        result = struct.unpack_from("<I", self.bytes, self.index)[0]
        self.index += 4
        return result

    def get_bool(self) -> bool:
        """
        Returns a boolean value from the byte array and advances the index.

        :return: A boolean value.
        """
        return self.get_uint() != 0

    def get_float(self) -> float:
        """
        Returns a floating-point number from the byte array and advances the index.

        :return: A floating-point number.
        """
        result = struct.unpack_from("<f", self.bytes, self.index)[0]
        self.index += 4
        return result



# -*- coding: utf-8 -*-
"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages, Extension

# To use a consistent encoding
from codecs import open
from os import path
import os
from glob import glob

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

reqs = []


setup_args = {
    "name": "pyhfst",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    "version": "1.3.0",
    "description": "A pure Python implementation of HFST for using HFST optimized lookup transducers (with or without weights)",
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    # The project's main homepage.
    "url": "https://github.com/Rootroo-ltd/pyhfst",
    # Author details
    "author": "Khalid Alnajjar and Mika Hämäläinen",
    "author_email": "hello@rootroo.com",
    "zip_safe": False,
    # Choose your license
    "license": "Apache-2.0",
    # See https://pypi.python.org/pypi?%3Aaction':list_classifiers
    "classifiers": [
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Text Processing :: Linguistic",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
    ],
    # What does your project relate to?
    "keywords": "hfst, fst, optimized lookup, hfstol, transducers",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    "packages": ["c_pyhfst", "pyhfst"],
    "package_dir": {"pyhfst": "pyhfst", "c_pyhfst": "c_pyhfst"},
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    "install_requires": reqs,
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    "extras_require": {},
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    "package_data": {
        "c_pyhfst": [
            x.split("/")[-1]
            for x in list(glob("./c_pyhfst/*.pyx"))
            + list(glob("./c_pyhfst/*.pxd"))
            + list(glob("./c_pyhfst/*.h"))
        ]
    },
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    "data_files": [],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    "entry_points": {},
    "project_urls": {  # Optional
        "Bug Reports": "https://github.com/Rootroo-ltd/pyhfst/issues",
        "Developer": "https://rootroo.com/",
    },
}

try:
    from Cython.Build import cythonize

    setup(ext_modules=cythonize("c_pyhfst/*.pyx", language_level=3), **setup_args)
except Exception as e:
    setup(**setup_args)



# cython: language_level=3
from libc.stdint cimport uint32_t, uint16_t, int32_t
import io
from .byte_array cimport ByteArray
cimport cython

cdef class IndexTable:
    def __init__(self, input_stream: io.BytesIO, cython.longlong indices_count):
        b = ByteArray(indices_count * 6)
        input_stream.readinto(b.bytes)

        self._size = indices_count
        self.ti_input_symbols = array.array('H', [])
        array.resize(self.ti_input_symbols, self._size)

        self.ti_targets = array.array('I', [])
        array.resize(self.ti_targets, self._size)

        for i in range(indices_count):
            self.ti_input_symbols[i] = b.get_ushort()
            self.ti_targets[i] = b.get_uint()

    cpdef uint16_t get_input(self, int i):
        i = i % self._size
        return self.ti_input_symbols[i]

    cpdef uint32_t get_target(self, int i):
        i = i % self._size
        return self.ti_targets[i]

    cpdef bint is_final(self, int i):
        i = i % self._size
        return (self.ti_input_symbols[i] == NO_SYMBOL_NUMBER and self.ti_targets[i] != NO_TABLE_INDEX)

    cpdef float get_final_weight(self, int i):
        i = i % self._size
        return float(self.ti_targets[i])


cdef class TransitionTable:
    def __init__(self, input_stream: io.BytesIO, cython.longlong transition_count, bint is_weighted=True):
        block_size = 12 if is_weighted else 8
        b = ByteArray(transition_count * block_size)
        input_stream.readinto(b.bytes)

        self._size = transition_count

        self.is_weighted = is_weighted

        self.ti_input_symbols = array.array('H', [])
        array.resize(self.ti_input_symbols, self._size)

        self.ti_output_symbols = array.array('H', [])
        array.resize(self.ti_output_symbols, self._size)

        self.ti_targets = array.array('I', [])
        array.resize(self.ti_targets, self._size)

        self.ti_weights = array.array('f', [])
        array.resize(self.ti_weights, self._size)

        for i in range(transition_count):
            self.ti_input_symbols[i] = b.get_ushort()
            self.ti_output_symbols[i] = b.get_ushort()
            self.ti_targets[i] = b.get_uint()
            if self.is_weighted:
                self.ti_weights[i] = b.get_float()

    cpdef uint16_t get_input(self, int pos):
        pos = pos % self._size
        return self.ti_input_symbols[pos]

    cpdef uint16_t get_output(self, int pos):
        pos = pos % self._size
        return self.ti_output_symbols[pos]

    cpdef uint32_t get_target(self, int pos):
        pos = pos % self._size
        return self.ti_targets[pos]

    cpdef float get_weight(self, int pos):
        pos = pos % self._size
        if not self.is_weighted:
            raise Exception("Getting weights of unweighted FST.")
        return self.ti_weights[pos]

    cpdef bint is_final(self, int pos):
        pos = pos % self._size
        return (self.ti_input_symbols[pos] == NO_SYMBOL_NUMBER and
                self.ti_output_symbols[pos] == NO_SYMBOL_NUMBER and
                self.ti_targets[pos] == 1)

    cpdef int size(self):
        return self._size

cdef class State:
    def __init__(self, str input, parent):
        self.parent = parent
        self.state_stack = []
        neutral = [0] * parent.alphabet.features
        self.state_stack.append(neutral)
        self.output_string = [NO_SYMBOL_NUMBER] * 1000
        self.input_string = self.find_key(input)
        self.output_pointer = 0
        self.input_pointer = 0
        self.current_weight = 0.0
        self.display_vector = []

    cdef list find_key(self, str index_string):
        output = list()
        i = 0

        while i < len(index_string):
            match_found = False
            for length in range(len(index_string) - i, 0, -1):
                substr = index_string[i : i + length]
                map_pointer = self.parent.symbol_map
                for char in substr:
                    if char in map_pointer:
                        map_pointer = map_pointer[char]
                    else:
                        break
                else:
                    if None in map_pointer:
                        output.append(map_pointer[None])
                        i += length - 1
                        match_found = True
                    break
            if not match_found:
                output.append(NO_SYMBOL_NUMBER)
            i += 1

        output.append(NO_SYMBOL_NUMBER)

        return output

cdef class Result:
    def __init__(self, list symbols, float weight):
        self.symbols = symbols
        self.weight = weight
    cpdef list get_symbols(self):
        return self.symbols

    cpdef float get_weight(self):
        return self.weight

    def __str__(self):
        return ''.join(self.symbols) + ": " + str(self.weight)



# cython: language_level=3
from .flag_diacritic_operator cimport FlagDiacriticOperator
from .flag_diacritic_operation cimport FlagDiacriticOperation
import io

cdef class TransducerAlphabet:
    def __init__(self, charstream: io.BytesIO, int number_of_symbols):
        """
        Initializes the TransducerAlphabet instance.

        :param filename: The file containing the alphabet data.
        :param number_of_symbols: The number of symbols in the alphabet.
        """
        self.keyTable = []
        self.operations = {}
        feature_bucket = {}
        value_bucket = {}
        self.features = 0
        self.values = 1
        value_bucket[""] = 0  # neutral value
        cdef str ops_index = "PNRDCU"

        cdef bytearray chars = bytearray()
        cdef str ustring
        cdef int charindex = 0
        cdef FlagDiacriticOperator op
        cdef str vals, feats, ops

        for _ in range(number_of_symbols):
            charindex = 0
            if len(chars) == charindex:
                chars.append(charstream.read(1)[0])
            else:
                chars[charindex] = charstream.read(1)[0]
            while chars[charindex] != 0:
                charindex += 1
                if len(chars) == charindex:
                    chars.append(charstream.read(1)[0])
                else:
                    chars[charindex] = charstream.read(1)[0]
            ustring = chars[:charindex].decode("utf-8")

            if (
                len(ustring) > 5
                and ustring[0] == ord("@")
                and ustring[-1] == ord("@")
                and ustring[2] == ord(".")
            ):  # flag diacritic identified
                parts = ustring[1:-1].split(".")
                if len(parts) < 2:
                    self.keyTable.append("")
                    continue
                ops, feats, *remainder = parts
                vals = remainder[0] if remainder else ""
                
                op = ops_index.index(ops)
                

                if op is None:  # Not a valid operator, ignore the operation
                    self.keyTable.append("")
                    continue

                if feats not in feature_bucket:
                    feature_bucket[feats] = self.features
                    self.features += 1
                if vals not in value_bucket:
                    value_bucket[vals] = self.values
                    self.values += 1
                self.operations[len(self.keyTable)] = FlagDiacriticOperation(
                    op, feature_bucket[feats], value_bucket[vals]
                )
                self.keyTable.append("")
                continue
            self.keyTable.append(ustring)
        self.keyTable[0] = ""  # epsilon is zero



import io
from typing import List, Dict
from .flag_diacritic_operation import FlagDiacriticOperator


class FlagDiacriticOperation:

    def __init__(self, operation, feat, val):
        self.op = operation
        self.feature = feat
        self.value = val


class TransducerAlphabet:
    def __init__(self, charstream: io.BytesIO, number_of_symbols: int) -> None:
        """
        Initializes the TransducerAlphabet instance.

        :param charstream: A byte stream containing the alphabet data.
        :param number_of_symbols: The number of symbols in the alphabet.
        """
        self.keyTable: List[str] = []
        self.operations: Dict[int, FlagDiacriticOperation] = {}
        feature_bucket: Dict[str, int] = {}
        value_bucket: Dict[str, int] = {}
        self.features = 0
        values = 1
        value_bucket[""] = 0  # neutral value

        chars = bytearray()
        for _ in range(number_of_symbols):
            charindex = 0
            if len(chars) == charindex:
                chars.append(charstream.read(1)[0])
            else:
                chars[charindex] = charstream.read(1)[0]
            while chars[charindex] != 0:
                charindex += 1
                if len(chars) == charindex:
                    chars.append(charstream.read(1)[0])
                else:
                    chars[charindex] = charstream.read(1)[0]
            ustring = chars[:charindex].decode("utf-8")

            if (
                len(ustring) > 5
                and ustring[0] == "@"
                and ustring[-1] == "@"
                and ustring[2] == "."
            ):  # flag diacritic identified
                op: FlagDiacriticOperator
                parts = ustring[1:-1].split(".")
                # Not a flag diacritic after all, ignore it
                if len(parts) < 2:
                    self.keyTable.append("")
                    continue
                ops, feats, *remainder = parts
                vals = remainder[0] if remainder else ""

                op = FlagDiacriticOperator[ops]

                if op is None:  # Not a valid operator, ignore the operation
                    self.keyTable.append("")
                    continue

                if vals not in value_bucket:
                    value_bucket[vals] = values
                    values += 1
                if feats not in feature_bucket:
                    feature_bucket[feats] = self.features
                    self.features += 1
                self.operations[len(self.keyTable)] = FlagDiacriticOperation(
                    op, feature_bucket[feats], value_bucket[vals]
                )
                self.keyTable.append("")
                continue
            self.keyTable.append(ustring)
        self.keyTable[0] = ""  # epsilon is zero



# cython: language_level=3

from libc.stdint cimport uint8_t, uint16_t, uint32_t

cdef class ByteArray:

    cdef bytearray bytes
    cdef int size
    cdef int index

    cpdef public int get_size(self)
    cpdef public bytes get(self, int i)
    cpdef public bytearray get_bytes(self)
    cpdef public uint8_t get_ubyte(self)
    cpdef public uint16_t get_ushort(self)
    cpdef public uint32_t get_uint(self)
    cpdef public bint get_bool(self)
    cpdef public float get_float(self)
